

from musiclang import Score
from transformers import StoppingCriteria, StoppingCriteriaList
import time
from musiclang import Chord
from fractions import Fraction as frac

class StoppingCriteriaSub(StoppingCriteria):

    def __init__(self, stops=[], encounters=1):
        super().__init__()
        self.stops = stops
        self.ENCOUNTERS = encounters

    def __call__(self, input_ids, scores, **kwargs):
        stop_count = 0
        for stop in self.stops:
            stop_count += (stop == input_ids[0]).sum().item()

        if stop_count >= self.ENCOUNTERS:
            return True
        return False


def extract_max_density_instruments(score):
    density = score.extract_densities()
    max_density_instrument_name = max(density, key=density.get).split('__')[0]
    fixed_parts = [ins for ins in score.instruments if ins.split('__')[0] ==max_density_instrument_name]
    return fixed_parts

def predict_melody(score_prompt, instrument, model, tokenizer, prompt_size=1024, context_size=2048, temperature=0.9, **kwargs):
    chord_duration = score_prompt.chords[0].duration
    inputs = tokenizer.tokenize_to_ids(score_prompt)[:-1]
    score, inputs = predict_melody_from_inputs(inputs, chord_duration, instrument,
                                               model, tokenizer, prompt_size=prompt_size,
                                                context_size=context_size, temperature=temperature, **kwargs
                                               )
    return score


def predict_melody_from_inputs(inputs, chord_duration, instrument, model, tokenizer, density=None,
                               octave=None, amplitude=None,
                               prompt_size=1024, context_size=2048, temperature=0.9, **kwargs):
    # Add density token
    if density is not None and tokenizer.dict['options'].get('density_token', False):
        if isinstance(density, str):
            density_str = density
        else:
            density_str = tokenizer.density_to_density_str(density)
        density_token = tokenizer['DENSITY__' + str(density_str)]
        inputs += [density_token]
    if octave is not None and tokenizer.dict['options'].get('average_octave_token', False):
        octave_token = tokenizer['AVERAGE_OCTAVE__' + str(octave)]
        inputs += [octave_token]
    if amplitude is not None and tokenizer.dict['options'].get('amplitude_token', False):
        amplitude_token = tokenizer['AMPLITUDE__' + str(amplitude)]
        inputs += [amplitude_token]

    nb = context_size - prompt_size
    end_token = tokenizer['END']
    generation_config = dict(
        max_new_tokens=nb,  # extends samples by nb tokens
        min_new_tokens=0,
        do_sample=True,  # but sample instead
        temperature=temperature,
        pad_token_id=end_token,
    )

    if tokenizer.dict['options'].get('melody_end_token', False):
        stop_words_ids = [tokenizer['MELODY_END']]
    else:
        stop_words_ids_instruments = [val for i, val in tokenizer.tokenizer.vocab.items() if
                                      i.startswith('INSTRUMENT_NAME__') or i.startswith('INSTRUMENT_PART__')]
        stop_words_ids_chords = [val for i, val in tokenizer.tokenizer.vocab.items() if i.startswith('CHORD_DEGREE__')]
        stop_words_ids = stop_words_ids_instruments + stop_words_ids_chords

    nb = context_size - prompt_size
    instrument_name, part = instrument.split('__')
    instrument_id = tokenizer['INSTRUMENT_NAME__' + instrument_name]
    inputs.append(instrument_id)
    if tokenizer.dict['options'].get('voice_token', False):
        instrument_part = tokenizer['INSTRUMENT_PART__' + str(part)]
        inputs.append(instrument_part)
    # Load model
    score, inputs = model_pred_one_shot(inputs, nb, stop_words_ids, model,
                        tokenizer, chord_duration, generation_config, nb_stop_word=1, **kwargs)
    return score, inputs

def predict_with_fixed_instruments_no_prompt(instruments, chord_duration, model, tokenizer, prompt_size=1024, context_size=2048, temperature=0.9, n_chords=1,
                                   chords=None,
                                   **kwargs):

    inputs = []
    score = None
    for idx, chord in enumerate(chords):
        inputs += tokenizer.tokens_to_ids(tokenizer.tokenize_chord(chord.set_duration(chord_duration)))
        if tokenizer.dict['options'].get('next_chord_token', False) and idx < len(chords) - 1:
            inputs += tokenizer.tokens_to_ids(tokenizer.tokenize_next_chord(chords[idx + 1].set_duration(chord_duration)))

        for instrument in instruments:
            score, inputs = predict_melody_from_inputs(inputs, chord_duration, instrument, model, tokenizer,
                                                       prompt_size=prompt_size,
                                                       context_size=context_size, temperature=temperature, **kwargs)
    return score

def predict_with_template(template,
                          model, tokenizer,
                          prompt=None,
                          prompt_included_in_template=True,
                          prompt_size=1024, context_size=2048, temperature=0.9
                          ):
    """
    Predict with a template of instruments and chord value for each chord of the score.
    Allow the music to follow a template.
    Parameters
    ----------
    template: dict
    chords: list of Chord
    model: transformers.GPT2LMHeadModel
    tokenizer: MusicLangTokenizer
    prompt: MusicLang.Score
    prompt_included_in_template: bool
    prompt_size: int
    context_size: int
    temperature: float

    Returns
    -------
    score: MusicLang.Score

    """

    ts_num, ts_den = template['time_signature']
    time_signature = template['time_signature']
    chord_duration = frac(4 * ts_num, ts_den)
    tonality = template['tonality']
    inputs = []

    # If prompt add it to inputs
    if prompt is not None:
        inputs += tokenizer.tokenize_to_ids(prompt)[:-1]
        if prompt_included_in_template:
            template['chords'] = template['chords'][len(prompt.chords):]

    score = None
    for idx, bar in enumerate(template['chords']):
        chord = Chord.from_romantext(bar['chord'], tonality=tonality, time_signature=time_signature)
        instruments_for_chord = bar['orchestration']
        inputs += tokenizer.tokens_to_ids(tokenizer.tokenize_chord(chord.set_duration(chord_duration)))
        if tokenizer.dict['options'].get('next_chord_token', False) and idx < len(template['chords']) - 1:
            next_chord = Chord.from_romantext(template['chords'][idx + 1]['chord'],
                                              tonality=tonality, time_signature=time_signature)
            inputs += tokenizer.tokens_to_ids(
                tokenizer.tokenize_next_chord(next_chord.to_chord().set_duration(chord_duration)))

        for instrument in instruments_for_chord:
            density = instrument['density']
            octave = instrument['octave']
            amplitude = instrument['amplitude']
            instrument = instrument['instrument_name'] + '__' + str(instrument['instrument_voice'])
            score, inputs = predict_melody_from_inputs(inputs, chord_duration, instrument, model, tokenizer,
                                                       density=density,
                                                       octave=octave,
                                                       amplitude=amplitude,
                                                       prompt_size=prompt_size,
                                                       context_size=context_size, temperature=temperature)
    return score


def predict_chords(model, tokenizer, nb_chords=4, prompt=None, temperature=0.9):

    if prompt is None:
        nb_chord_init = 0
        prompt = "CHORD_CHANGE"
    else:
        nb_chord_init = len(prompt.chords)
        prompt = " ".join(tokenizer.tokenize_chords(prompt))

    ids = tokenizer(prompt, add_special_tokens=False, return_tensors='pt').input_ids
    nb_tokens = (nb_chords + 1 + nb_chord_init) * 6
    result = model.generate(ids, do_sample=True, temperature=temperature,
                            #max_new_tokens=nb_tokens
                        max_length= nb_tokens, min_length=nb_tokens
                            )[0]
    seq = tokenizer.decode(result)
    score = tokenizer.untokenize(prompt + ' ' + seq)
    score = score[:nb_chords]

    return score


def predict_with_fixed_instruments(score_prompt, model, tokenizer, prompt_size=1024, context_size=2048, temperature=0.9, n_chords=1,
                                   chords=None,
                                   **kwargs):
    instruments = score_prompt[-1:].instruments  # Use last instruments by default
    chord_duration = score_prompt.chords[0].duration
    score = score_prompt

    inputs = tokenizer.tokenize_to_ids(score)[:-1]
    if chords is None:
        chords = [score_prompt[-1].to_chord() for i in range(n_chords)]

    for chord in chords:
        # FIXME : New chord not possible, to fix with next tokenizer
        inputs += tokenizer.tokenize_to_ids(chord.set_duration(chord_duration))[:6]
        for instrument in instruments:
            score, inputs = predict_melody_from_inputs(inputs, chord_duration, instrument, model, tokenizer, prompt_size=prompt_size,
                                   context_size=context_size, temperature=temperature, **kwargs)
        # Add chord tokens
    return score


def model_pred_one_shot(inputs, nb, stop_words_ids, model, tokenizer, chord_duration, generation_config,
                        nb_stop_word=1,
                        **kwargs):
    import torch

    from musiclang.transform import create_counterpoint_on_score
    prompt = inputs[-nb:]
    remaining = len(inputs) - len(prompt)
    inputs = inputs[:remaining]
    # Load model
    prompt = torch.tensor([prompt], dtype=torch.long)
    nb_encounters = len([i for i in prompt[0] if i in stop_words_ids])
    stopping_criteria = StoppingCriteriaList(
        [StoppingCriteriaSub(stops=stop_words_ids, encounters=nb_encounters + nb_stop_word)])

    tokens = model.generate(prompt, stopping_criteria=stopping_criteria, **generation_config)[0]  # (N,T)
    new_ids = tokens.tolist()
    inputs += new_ids

    # Correct inputs
    score = tokenizer.ids_to_score(inputs[:-1])  # Remove stop token
    # score = score.arrange_chords_duration(chord_duration)
    # drums = score.get_instrument_names(['drums_0'])
    # score_without_drums = score.normalize_instruments().remove_drums()
    #
    # if len(score_without_drums.instruments) > 0:
    #     score = score_without_drums
    #     fixed_parts = extract_max_density_instruments(score)
    #     #score = create_counterpoint_on_score(score, fixed_parts=fixed_parts)
    #     # Reproject the drums
    #     if len(drums.instruments) > 0:
    #         score = score.project_on_score(drums, voice_leading=False, keep_score=True)
    # score = score.remove_silenced_instruments()

    inputs = tokenizer.tokenize_to_ids(score, include_end=False)
    return score, inputs


def predict(model, tokenizer, prompt=None,
            chord_duration=4,
            prompt_size=1024, context_size=2048, nb_chords=1, temperature=0.9, **kwargs):
    """
    Predict continuation of a score

    Parameters
    ----------
    The score to continue
    model: transformers.GPT2LMHeadModel
    The model to use for prediction
    tokenizer: MusicLangTokenizer
    The tokenizer to use for prediction
    prompt: musiclang.Score
    The score to continue
    chord_duration: int
    The duration of the chords to predict in number of quarter notes
    prompt_size: int
    The size of the prompt to use for prediction
    context_size: int
    The size of the context to use for prediction
    nb_chords: int
    The number of chords to predict
    temperature: float
    The temperature to use for prediction
    kwargs

    Returns
    -------

    """
    stop_words_ids = [tokenizer['END'], tokenizer['CHORD_CHANGE']]
    nb = context_size - prompt_size
    generation_config = dict(
        max_new_tokens=nb,  # extends samples by nb tokens
        min_new_tokens=0,
        min_length=0,
        do_sample=True,  # but sample instead
        temperature=temperature,
        pad_token_id=tokenizer['[PAD]'] if '[PAD]' in tokenizer.tokenizer.vocab else tokenizer['END'],
    )


    if prompt:
        chord_duration = prompt.chords[0].duration
        inputs = tokenizer.tokenize_to_ids(prompt, include_end=False)
    else:
        inputs = []
    score = None
    for i in range(nb_chords):
        inputs = inputs + [tokenizer['CHORD_CHANGE']]
        score, inputs = model_pred_one_shot(inputs, nb, stop_words_ids, model,
                                            tokenizer, chord_duration, generation_config, nb_stop_word=1, **kwargs)

    return score



from musiclang_predict import MusicLangTokenizer
from musiclang.library import *


default_options = {
    'chord_change_token': True,
    'melody_end_token': True,
    'chord_duration_token': True,
    'density_token': True,
    'chord_extension_token': True,
    'next_chord_token': True,
    'next_chord_duration_token': True,
    'will_end_token': True,
    'dissonance_token': True,
    'amplitude_token': True,
    'average_octave_token': True,
    'voice_token': True,
    'random_instrument_permutation': False
}

def test_tokenizer():

    pass



def test_cut_bar():
    pass

def test_add_bar():
    pass


from .tokenizers import MusicLangTokenizer, midi_file_to_template, score_to_template, MusicLangBPETokenizer
from .predictors import predict, predict_melody, predict_with_fixed_instruments, \
    predict_with_fixed_instruments, predict_with_fixed_instruments_no_prompt, predict_with_template, predict_chords
from .predict import MusicLangPredictor


__all__ = ['predict_with_fixed_instruments_no_prompt', 'predict_with_template',
           'MusicLangTokenizer', 'predict', 'predict_melody', 'predict_with_fixed_instruments',
           'predict_with_fixed_instruments', 'midi_file_to_template', 'score_to_template',
           'predict_chords', 'MusicLangBPETokenizer', 'MusicLangPredictor'
           ]



from musiclang.library import *
import torch

TEST_CHORD = (I % I.M)


def get_nb_tokens_chord(tokenizer):
    return len(tokenizer(TEST_CHORD))



def predict_with_chords(chord_score, model, tokenizer, chord_duration=4, temperature=1.1, top_k=20, **model_kwargs):


    nb_tokens_chord = get_nb_tokens_chord(tokenizer)
    chord_score = chord_score.to_score().set_duration(chord_duration)

    stop_token = tokenizer.tokenizer(tokenizer.tokens_to_bytes(['CHORD_CHANGE']))['input_ids']
    # We generate one chord per one chord
    eos_token_id = int(stop_token[0])
    result = []

    for chord in chord_score:
        # We generate the chord
        chord_tokens = tokenizer(chord.to_score())[:nb_tokens_chord]
        result += chord_tokens
        # result += [67]

        new_ids = model.generate(torch.tensor([result]), do_sample=True,
                                 eos_token_id=eos_token_id,
                                 temperature=temperature, top_k=top_k,
                                 max_new_tokens=500,
                                 **model_kwargs)[0].tolist()[:-1]

        result = new_ids

    score = tokenizer.ids_to_score(result)
    return score




from transformers import AutoTokenizer



class MusicLangBPETokenizer:

    def __init__(self, tokenizer_path, pretokenizer_path):
        from musiclang_predict import MusicLangTokenizer

        self.pretokenizer = MusicLangTokenizer(pretokenizer_path)
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)

    def tokens_to_bytes(self, tokens):
        """
        Convert a list of tokens to a string of bytes.
        :param tokens:
        :return:
        """
        return self.pretokenizer.tokens_to_bytes(tokens)

    def __call__(self, score, **kwargs):
        """
        Tokenize the input text.
        :param text: input text.
        :return: list of tokens.
        """
        tokens = self.pretokenizer.tokenize(score)
        bytes_ = self.tokens_to_bytes(tokens)

        return self.tokenizer(bytes_, **kwargs).input_ids

    def ids_to_text(self, ids):
        bytes_ = self.tokenizer.decode(ids).replace(' ', '')
        text = self.pretokenizer.bytes_to_tokens(bytes_)
        return text


    def untokenize(self, ids):
        return self.ids_to_score(ids)

    def ids_to_score(self, ids):
        """
        Convert a list of tokens to a string of bytes.
        :param tokens:
        :return:
        """
        bytes_ = self.tokenizer.decode(ids).replace(' ', '')
        text = self.pretokenizer.bytes_to_tokens(bytes_)
        return self.pretokenizer.untokenize(text)


import ctypes
import os

current_file_path = os.path.dirname(os.path.abspath(__file__))



def load_library():
    # Assuming your package structure follows a certain pattern
    lib_name = "librun.so" if os.name != "nt" else "run.dll"
    lib_path = os.path.join(current_file_path, "c", lib_name)
    return ctypes.CDLL(lib_path)

def run_transformer_model(checkpoint_path, tokenizer_path, temperature=1.0, topp=1.0, rng_seed=0, steps=256,
                          prompt=None, mode="generate", system_prompt=None,
                          stop_char="_"
                          ):

    lib = load_library()

    # Define the return type and argument types for run_model
    lib.run_model.restype = ctypes.c_void_p  # Use void pointer for the returned string
    lib.run_model.argtypes = [
        ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char, ctypes.c_float, ctypes.c_float,
        ctypes.c_ulonglong, ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p
    ]

    # Define the argument types for free_returned_string
    lib.free_returned_string.restype = None
    lib.free_returned_string.argtypes = [ctypes.c_void_p]

    # Convert string parameters to bytes, if they are not None
    if prompt is not None:
        prompt = prompt.encode('utf-8')
    if mode is not None:
        mode = mode.encode('utf-8')
    if system_prompt is not None:
        system_prompt = system_prompt.encode('utf-8')
    if stop_char is not None:
        stop_char = stop_char.encode('utf-8')
    if stop_char is None:
        stop_char = " ".encode('utf-8')
    # Call the function
    result_ptr = lib.run_model(
        checkpoint_path.encode('utf-8'), tokenizer_path.encode('utf-8'), stop_char, temperature, topp, rng_seed,
        steps, prompt, mode, system_prompt
    )

    # Convert the result to a Python string
    result_str = ctypes.cast(result_ptr, ctypes.c_char_p).value.decode('utf-8')

    # Free the returned string after use
    lib.free_returned_string(result_ptr)

    return result_str


def run_for_n_bars(n_bars, checkpoint_path, tokenizer_path, temperature=1.0, topp=1.0, rng_seed=0, steps=256,
                          prompt=None, mode="generate", system_prompt=None,
                          stop_char="_"):

    for i in range(n_bars):
        generated_text = run_transformer_model(
            checkpoint_path=checkpoint_path,
            tokenizer_path=tokenizer_path,
            temperature=temperature,
            topp=topp,
            rng_seed=rng_seed,
            steps=steps,
            prompt=prompt,
            mode=mode,
            system_prompt=system_prompt,
            stop_char=stop_char
        )
        print(generated_text)
        prompt = generated_text
    return prompt





from musiclang import Score
from musiclang.library import *

from musiclang_predict.chelpers import run_for_n_bars, run_transformer_model
import os
import huggingface_hub
from huggingface_hub import hf_hub_download
from musiclang_predict import MusicLangTokenizer

STOP_CHAR = None
CHORD_CHANGE_CHAR = "_"

TEST_CHORD = (I % I.M)


def get_nb_tokens_chord(tokenizer):
    return len(tokenizer.tokenize_to_bytes(TEST_CHORD))

MIDI_EXTENSIONS = ['mid', 'midi', 'MID', 'MIDI']
XML_EXTENSIONS = ['xml', 'mxl', 'musicxml', 'XML', 'MXL', 'MUSICXML']

class MusicLangPredictor:

    def __init__(self, path, tokenizer_file="tokenizer.bin", model_file="model.bin"):
        self.path = path
        self.tokenizer_path = hf_hub_download(repo_id=self.path, filename=tokenizer_file)
        self.model_path = hf_hub_download(repo_id=self.path, filename=model_file)
        self.pretokenizer = MusicLangTokenizer(self.path)
        self._nb_tokens_chord = get_nb_tokens_chord(self.pretokenizer)

    def parse_score(self, score, prompt_chord_range=None):
        # Tokenize the score to bytes

        if isinstance(score, str) and score.split('.')[-1] in MIDI_EXTENSIONS:
            # Load score
            from musiclang import Score
            score = Score.from_midi(score, chord_range=prompt_chord_range)
        elif isinstance(score, str) and score.split('.')[-1] in XML_EXTENSIONS:
            # Load score
            from musiclang import Score
            if prompt_chord_range is not None:
                raise ValueError("Sorry ... Chord range is not supported yet for XML files, convert it to midi first")
            score = Score.from_xml(score)
        score = self.pretokenizer.tokenize_to_bytes(score, self.pretokenizer) + CHORD_CHANGE_CHAR
        return score

    def predict(self, score=None, prompt_chord_range=None, nb_tokens: int = 256, temperature=0.9, topp=1.0, rng_seed=0):
        """
        Generate a score from a prompt
        :param score: (Optional) MusicLang Score, midi or xml file, default None
        The prompt used to continue the generation on
        :param prompt_chord_range: (Optional) tuple (int, int), default None
        Chord range to use for the prompt
        :param nb_tokens: (Optional) int, default 256
        Number of tokens to generate
        :param temperature: (Optional) float, default 0.9
        Temperature to use for the generation
        :param topp: (Optional) float, default 1.0
        Top-p to use for the generation
        :param rng_seed: (Optional) int, default 0
        Random seed to use for the generation. Use 0 for no seed
        :return: MusicLang Score
        The generated score
        """
        if score is not None:
            score = self.parse_score(score, prompt_chord_range)

        generated_text = run_transformer_model(
            checkpoint_path=self.model_path,
            tokenizer_path=self.tokenizer_path,
            temperature=temperature,
            topp=topp,
            rng_seed=rng_seed,
            steps=nb_tokens,
            prompt=score,
            mode="generate",
            system_prompt=None,
            stop_char=STOP_CHAR)

        # Untokenize to score
        score = self.pretokenizer.untokenize_from_bytes(generated_text)
        return score

    def ts_to_duration(self, ts):
        from fractions import Fraction
        return Fraction(4 * ts[0], ts[1])

    @property
    def nb_tokens_chord(self):
        return self._nb_tokens_chord

    def chord_to_tokens(self, chord):
        return self.pretokenizer.tokenize_to_bytes(chord, self.pretokenizer)[1:self.nb_tokens_chord]

    def predict_chords(self, chords: str, time_signature=(4, 4), score=None, prompt_chord_range=None, nb_tokens: int = 4096, temperature=0.9, topp=1.0, rng_seed=0):

        chord_duration = self.ts_to_duration(time_signature)
        chords = Score.from_chord_repr(chords)
        chords = chords.set_duration(chord_duration)
        chord_tokens = [self.chord_to_tokens(chord) for chord in chords]

        if score is not None:
            score = self.parse_score(score, prompt_chord_range)
        else:
            score = CHORD_CHANGE_CHAR

        for chord in chord_tokens:

            score += chord

            generated_text = run_transformer_model(
                checkpoint_path=self.model_path,
                tokenizer_path=self.tokenizer_path,
                temperature=temperature,
                topp=topp,
                rng_seed=rng_seed,
                steps=nb_tokens,
                prompt=score,
                mode="generate",
                system_prompt=None,
                stop_char=CHORD_CHANGE_CHAR)

            score = generated_text

        score = self.pretokenizer.untokenize_from_bytes(score)

        return score




from musiclang import Score

def midi_file_to_template(midi_file, chord_range=None, max_instruments=8, quantization=(4, 3)):
    """
    Extract a song template from a midi file. It will extract the chord progression, the orchestration,
    The average density, the average amplitude, the average octave for each instrument of each bar.

    It will also extract metadata about the soundtrack like the tonality, the tempo and the time signature.
    :param midi_file: str, path to midi file
    :param chord_range: tuple, range of chords to extract (start, end) (default=None)
    :param max_instruments: int, maximum number of instruments to extract (default=8)
    :return: dict, template
    """
    score_prompt = Score.from_midi(midi_file, quantization=quantization, chord_range=chord_range)
    return score_to_template(score_prompt, max_instruments=max_instruments)


def score_to_template(score, max_instruments=8):
    """
    Extract a song template from a musiclang score. It will extract the chord progression, the orchestration,
    The average density, the average amplitude, the average octave for each instrument of each bar.

    It will also extract metadata about the soundtrack like the tonality, the tempo and the time signature.
    :param score: str, path to midi file
    :param chord_range: tuple, range of chords to extract (start, end) (default=None)
    :param max_instruments: int, maximum number of instruments to extract (default=8)
    :return: dict, template
    """
    score_prompt = score.to_score()
    densities_per_chords = [chord.to_score().extract_densities() for chord in score_prompt]
    amplitudes_per_chords = [chord.to_score().extract_mean_amplitudes() for chord in score_prompt]
    octaves_per_chords = [chord.to_score().extract_mean_octaves() for chord in score_prompt]
    tonality, chord_list = score_prompt.to_romantext_chord_list()
    data_chords = [
        {
            "orchestration": [{'instrument_name': key.split('__')[0],
                               'instrument_voice': int(key.split('__')[1]),
                               'amplitude': amplitudes_per_chords[idx][key],
                               'octave': octaves_per_chords[idx][key],
                               'density': float(val)}
                              for key, val in densities_per_chord.items()],
            "chord": chord_list[idx]
        }
        for idx, densities_per_chord in enumerate(densities_per_chords)

    ]

    for data_chord in data_chords:
        data_chord['orchestration'] = data_chord['orchestration'][:max_instruments]

    ts = score_prompt.config['time_signature']
    if len(ts) == 4:
        ts = ts[1], ts[2]
    data = {
        'tonality': tonality,
        'tempo': score_prompt.config['tempo'],
        'time_signature': ts,
        'chords': data_chords
    }
    return data



import copy

from musiclang import Score, Chord, Note, Melody, Tonality
import os
import tempfile
import os

import json
from fractions import Fraction as frac
import joblib
import numpy as np
from multiprocessing import Pool
import functools
from huggingface_hub import hf_hub_download
from tqdm import tqdm
import gc
from fractions import Fraction as frac

from tokenizers import Tokenizer, models, trainers

from transformers import AutoTokenizer, AutoModelForCausalLM, PreTrainedTokenizerFast
from tokenizers import Tokenizer
from tokenizers.models import WordLevel
from tokenizers.pre_tokenizers import WhitespaceSplit
from tokenizers.trainers import WordLevelTrainer


from musiclang_predict.tokenizers.bpe_iterator import BPEIterator

NOTE_DURATION_MAX_DENOMINATOR = 8
CHORD_DURATION_MAX_DENOMINATOR = 4
BASE_CHAR_ID = 33
TOKENIZER_CONFIG_BASE = {
  "model_max_len": 4096,
  "model_max_length": 4096,
  "pad_token": "[PAD]",
  "pad_token_id": 3,
  "eos_token": "[END]",
  "eos_token_id": 45,
  "bos_token": "[START]",
  "bos_token_id": 5,
  "mask_token": "[MASK]",
  "mask_token_id": 4,
  "unk_token": "[UNK]",
  "unk_token_id": 0,
  "cls_token": "[CLS]",
  "cls_token_id": 1,
  "sep_token": "[SEP]",
  "seo_token_id": 2,
  "add_prefix_space": False,
  "tokenizer_class": "PreTrainedTokenizer",
  "type": "Split"
}


default_options = {
    'chord_change_token': True,
    'melody_end_token': True,
    'chord_duration_token': True,
    'density_token': True,
    'chord_extension_token': True,
    'next_chord_token': True,
    'next_chord_duration_token': True,
    'will_end_token': True,
    'dissonance_token': True,
    'amplitude_token': True,
    'average_octave_token': True,
    'voice_token': False,
    'random_instrument_permutation': True,
    'end_token': True,
    'silence_continuation_eco': True
}
class MusicLangTokenizer:
    """
    Convert a score into a list of tokens
    """

    SCORE_START = 'SCORE_START'
    UNKNOWN = 'UNKNOWN'

    CHORD_DEGREE = 'CHORD_DEGREE'
    TONALITY_DEGREE = 'TONALITY_DEGREE'
    TONALITY_MODE = 'TONALITY_MODE'
    CHORD_OCTAVE = 'CHORD_OCTAVE'
    CHORD_DURATION_NUM = 'CHORD_DURATION_NUM'
    CHORD_DURATION_DEN = 'CHORD_DURATION_DEN'

    CHORD_EXTENSION = 'CHORD_EXTENSION'

    NEXT_CHORD_DEGREE = 'NEXT_CHORD_DEGREE'
    NEXT_TONALITY_DEGREE = 'NEXT_TONALITY_DEGREE'
    NEXT_TONALITY_MODE = 'NEXT_TONALITY_MODE'
    NEXT_CHORD_OCTAVE = 'NEXT_CHORD_OCTAVE'
    NEXT_CHORD_EXTENSION = 'NEXT_CHORD_EXTENSION'
    NEXT_CHORD_DURATION_NUM = 'NEXT_CHORD_DURATION_NUM'
    NEXT_CHORD_DURATION_DEN = 'NEXT_CHORD_DURATION_DEN'

    CHORD_CHANGE = 'CHORD_CHANGE'
    MELODY_END = 'MELODY_END'
    WILL_END = 'WILL_END'
    DISSONANCE = 'DISSONANCE'
    AMPLITUDE = 'AMPLITUDE'

    INSTRUMENT_NAME = 'INSTRUMENT_NAME'
    INSTRUMENT_PART = 'INSTRUMENT_PART'
    DENSITY = 'DENSITY'
    AVERAGE_OCTAVE = 'AVERAGE_OCTAVE'

    NOTE_TYPE = 'NOTE_TYPE'
    NOTE_VAL = 'NOTE_VAL'
    NOTE_OCTAVE = 'NOTE_OCTAVE'
    NOTE_AMP = 'NOTE_AMP'
    NOTE_DURATION_NUM = 'NOTE_DURATION_NUM'
    NOTE_DURATION_DEN = 'NOTE_DURATION_DEN'
    END = 'END'

    def __init__(self, tokenizer_path=None, options=None, hub_tokenizer_path='tokenizer-base.json'):
        self.dict = {}
        self.tokenizer = None
        self.denoms = [i for i in range(1, NOTE_DURATION_MAX_DENOMINATOR + 1)]
        if tokenizer_path is None:
            import warnings
            warnings.warn("No tokenizer_path provided. Using a new tokenizer. You probably should train it using 'train_tokenizer_from_token_files' method.")
        else:
            try:
                with open(tokenizer_path, 'r') as f:
                    self.dict = json.load(f)
            except Exception as e:
                tokenizer_path_hub = hf_hub_download(repo_id=tokenizer_path, filename=hub_tokenizer_path)
                with open(tokenizer_path_hub, 'r') as f:
                    self.dict = json.load(f)
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)
            self.denoms = self.get_avalaible_denominators()
        # Replace str to int for keys of id_to_token
        if options is not None:
            self.dict['options'] = options
        elif 'options' not in self.dict:
            self.dict['options'] = default_options

    @property
    def vocab_size(self):
        return len(self.tokenizer.vocab_size)

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.tokenizer.vocab[item]
        elif isinstance(item, int):
            return self.tokenizer.decode(item)
        else:
            raise ValueError(f"Invalid type {type(item)} for item {item}")

    def tokenize_midi_file(self, midi_file, quantization=8, fast_chord_inference=True, chord_range=None, interceptors=None):
        """
        Tokenize a single midi file and returns list of tokens
        Parameters
        ----------
        midi_file
        quantization
        fast_chord_inference
        chord_range

        Returns
        -------

        """
        score = Score.from_midi(midi_file, quantization=quantization, fast_chord_inference=fast_chord_inference, chord_range=chord_range)
        if interceptors is not None:
            for interceptor in interceptors:
                score = interceptor(score)
        tokens = self.tokenize(score)
        return tokens

    def tokenize_midi_files(self, midi_files, quantization=8, fast_chord_inference=True, chord_range=None):
        all_tokens = []
        for midi_file in midi_files:
            tokens = self.tokenize_midi_file(midi_file, quantization=quantization, fast_chord_inference=fast_chord_inference, chord_range=chord_range)
            all_tokens.append(tokens)
        return all_tokens


    def calculate_tokens_to_ids_dict(self, token_files):

        unique_tokens = set()
        for token_file in token_files:
            with open(token_file, 'r') as f:
                tokens = f.read().split('\n')
                unique_tokens.update(tokens)
        unique_tokens = list(sorted(list(unique_tokens)))
        dict = {token: idx for idx, token in enumerate(unique_tokens)}
        inv_dict = {idx: token for idx, token in enumerate(unique_tokens)}
        self.dict = {'options': self.dict['options']}


    def tokenize_from_file(self, path):
        """
        Tokenize a file and returns a list of tokens
        Parameters
        ----------
        path: str, path to file

        Returns
        -------
        tokens: List[str], list of tokens
        """
        with open(path, 'r') as f:
            tokens = f.read().split()
        return tokens

    def tokenize_sequence(self, seq):
        """
        Tokenize a sequence of tokens (A pandas dataframe with appropriate columns
        Parameters
        ----------
        seq: pd.DataFrame

        Returns
        -------
        tokens: str
        """

        score = Score.from_sequence(seq)
        return self.tokenize(score)



    def train_tokenizer_from_token_files(self, token_files, output=None, hub_output=None, **kwargs):
        """
        Train a tokenizer from a list of token files.
        It will also save a tokenizer-base.json file with the options used to train the tokenizer
        (needed for tokenization from musiclang language)

        Make sure you have logged in to huggingface using `huggingface-cli login` before training and pushing to the hub
        Parameters
        ----------
        token_files:
        output_tokenizer: Path to save the tokenizer
        output: Path to save the tokenizer and the config file (Either output or hub_output must be not None)
        hub_output: Path to save the tokenizer and the config file in the huggingface hub

        Returns
        -------
        tokenizer: Tokenizer
        """


        def train_tokenizer(data_files, vocab_size=30_000,
                            special_tokens=["[UNK]", "[CLS]", "[SEP]", "[PAD]", "[MASK]", "[START]", "[END]"]):
            # Create a tokenizer
            tokenizer = Tokenizer(WordLevel(unk_token="[UNK]"))
            # Pre-tokenizer to split the text into words
            tokenizer.pre_tokenizer = WhitespaceSplit()
            # tokenizer.pre_tokenizer = PreTokenizer.custom("whitespace_split", "regex", pattern=r"\s+")
            # Special tokens
            tokenizer.add_special_tokens(special_tokens)
            # Trainer
            trainer = WordLevelTrainer(
                # vocab_size=vocab_size,
                special_tokens=special_tokens,
            )

            # Train the tokenizer
            tokenizer.train(files=data_files, trainer=trainer)
            return tokenizer

        # Example usage
        data_files = token_files  # Replace with the path to your text file
        tokenizer = train_tokenizer(data_files)
        # Save the tokenizer to a temp directory
        with tempfile.TemporaryDirectory() as tmpdirname:
            tokenizer.save(os.path.join(tmpdirname, "tokenizer.json"))

            # Save TOKENIZER_BASE as tokenizer_config.json in directory
            options = copy.deepcopy(TOKENIZER_CONFIG_BASE)
            for key, value in kwargs.items():
                options[key] = value

            with open(os.path.join(tmpdirname, 'tokenizer_config.json'), 'w') as f:
                json.dump(options, f, indent=4)
            # Reload tokenizer to push it or save it
            tokenizer = AutoTokenizer.from_pretrained(tmpdirname)
            if output is not None:
                tokenizer.save_pretrained(output)
                # Save base options
                with open(os.path.join(output, 'tokenizer-base.json'), 'w') as f:
                    json.dump({"options": self.dict['options']}, f, indent=4)
            if hub_output is not None:
                tokenizer.push_to_hub(hub_output)
                # Push base options to hub (1. save to a tempfile, 2. then use hf api to push to hub)
                with tempfile.TemporaryDirectory() as tmpdirname:
                    with open(os.path.join(tmpdirname, 'tokenizer-base.json'), 'w') as f:
                        json.dump({"options":  self.dict['options']}, f, indent=4)
                    from huggingface_hub import HfApi
                    hf_api = HfApi()
                    hf_api.upload_file(
                        path_or_fileobj=os.path.join(tmpdirname, 'tokenizer-base.json'),
                        path_in_repo="tokenizer-base.json",
                        repo_id=hub_output,
                        repo_type="model",
                    )

            if output is None and hub_output is None:
                # Raise ONLY a warning
                print("WARNING: hub_output is None, tokenizer not pushed to hub")

        return tokenizer

    def __call__(self, score, **kwargs):
        tokens = " ".join(self.tokenize(score))
        return self.tokenizer(tokens, **kwargs)

    def decode(self, *args, **kwargs):
        return self.tokenizer.decode(*args, **kwargs)

    def tokenize_chords(self, score):
        return self.tokenize(score, only_chords=True)

    def ids_to_tokens(self, ids):
        """
        Convert a list of token ids to a list of tokens
        Parameters
        ----------
        ids

        Returns
        -------

        """
        return self.tokenizer.convert_ids_to_tokens(ids)

    def ids_to_score(self, ids):
        """
        Convert a list of token ids to a score
        Parameters
        ----------
        ids

        Returns
        -------

        """
        tokens = self.ids_to_tokens(ids)
        return self.untokenize(tokens)

    def file_ids_to_score(self, path):
        """
        Convert a file with ids to a score
        Parameters
        ----------
        path: Path to numpy array with ids

        Returns
        -------

        """
        tokens_ids = joblib.load(path)
        tokens = self.ids_to_tokens(tokens_ids)

        return self.untokenize(tokens)


    def file_tokens_to_score(self, path):
        """
        Convert a file with tokens to a score
        Parameters
        ----------
        path: Path to file with tokens

        Returns
        -------

        """
        with open(path, 'r') as f:
            tokens = f.read().split('\n')
        return self.untokenize(tokens)

    def tokenize_to_ids(self, score, include_end=True):
        """
        Tokenize a score and returns a list of token ids
        Parameters
        ----------
        score: MusicLang.Score, score to tokenize
        include_end: bool, if True add the END token at the end of the list (default=True)

        Returns
        -------
        ids: List[int], list of token ids

        """
        tokens = self.tokenize(score)
        ids = self.tokens_to_ids(tokens)
        if not include_end:
            ids = ids[:-1]
        return ids

    def tokens_to_ids(self, tokens):
        """
        Convert a list of tokens to a list of token ids
        Parameters
        ----------
        tokens

        Returns
        -------

        """
        if self.tokenizer is None:
            raise ValueError("No tokens to ids available. You must train the tokenizer first using train_tokenizer_from_token_files method.")
        res_with_tokenizer = self.tokenizer(" ".join(tokens))

        return res_with_tokenizer['input_ids']


    def train_bpe(self, files_paths, output_dir=None, hub_path=None, vocab_size=30_000, type='sentence_piece'):
        """
        :param files_paths: list of str
        Tokens text files paths
        :param output_dir: Output of the BPE model, if None, a temporary directory will be created
        :param hub_path: str
        Path to the hub where to push the tokenizer
        :param vocab_size: int
        Number of tokens in the BPE model
        :param type: str (sentence_piece or bpe)
        :return:
        """


        from tokenizers import SentencePieceBPETokenizer
        from tokenizers import normalizers
        # Initialize the BPEIterator with your custom tokenizer and list of file paths
        control_tokens = self.get_control_tokens_bytes()
        bpe_iterator = BPEIterator(self, files_paths, control_tokens=control_tokens)

        # Initialize the Hugging Face tokenizer with a BPE model
        if type == 'sentence_piece':
            tokenizer = SentencePieceBPETokenizer(add_prefix_space = False)
            tokenizer.enable_padding(pad_id=1, pad_token="<pad>")
            tokenizer.normalizer = normalizers.Sequence([])
            tokenizer.train_from_iterator(bpe_iterator, special_tokens=['<unk>', '<pad>', '<bos>', '<eos>', '<mask>'], vocab_size=vocab_size, show_progress=True)
        elif type == 'bpe':
            tokenizer = Tokenizer(models.BPE())
            trainer = trainers.BpeTrainer(vocab_size=vocab_size, show_progress=True, max_token_length=32)
            tokenizer.train_from_iterator(bpe_iterator, trainer=trainer)
        else:
            raise ValueError(f"Invalid type {type}, must be 'sentence_piece' or 'bpe'")

        # Save the trained tokenizer
        file = "tokenizer.json"
        delete_temp_dir = False
        if output_dir is None:
            output_dir = tempfile.mkdtemp()
            delete_temp_dir = True

        tokenizer_file = os.path.join(output_dir, file)
        tokenizer.save(tokenizer_file)
        tokenizer = Tokenizer.from_file(tokenizer_file)
        tokenizer.model.save(output_dir)

        if hub_path is not None:
            tokenizer = PreTrainedTokenizerFast(tokenizer_file=tokenizer_file, model_max_length=self.tokenizer.model_max_length)
            tokenizer.push_to_hub(hub_path)

        # Delete the temporary directory
        if delete_temp_dir:
            import shutil
            shutil.rmtree(output_dir)

        # Push to hub

        return tokenizer


    def tokens_to_bytes(self, tokens, as_one_str=True):
        if not isinstance(tokens, str):
            tokens = " ".join(tokens)
        ids = self.tokenizer(tokens)['input_ids']
        bytes_ids = [chr(i + BASE_CHAR_ID) for i in ids]
        if as_one_str:
            return ''.join(bytes_ids)
        return bytes_ids


    def get_control_tokens(self):
        """
        Get all the tokens that are control tokens and that should not be part of BPE merges
        :return:
        """

        # Control tokens are all tokens that are not note tokens
        control_tokens = [voc for voc in self.tokenizer.vocab.keys() if not voc.startswith('NOTE_')]
        return control_tokens

    def get_control_tokens_bytes(self):
        """
        Get all the tokens that looks like control tokens
        :return:
        """
        # Control tokens are all tokens that are not note tokens
        control_tokens = [chr(i + BASE_CHAR_ID) for i in self.tokenizer.vocab.values() if not self.tokenizer.decode(i).startswith('NOTE_')]
        return control_tokens

    def bytes_to_tokens(self, bytes, to_str=True):
        ids = [ord(b) - BASE_CHAR_ID for b in bytes]
        ids = [id for id in ids if id >= 0]
        result = self.ids_to_tokens(ids)
        if to_str:
            return " ".join(result)
        return result

    def save(self, filepath):
        with open(filepath, 'w') as f:
            json.dump(self.dict, f, indent=4)

    def tokenize_to_files_single(self, midi_files, output_dir, quantization=8,
                          fast_chord_inference=True, chord_range=None, sep="\n",
                          allow_error=True
                          ):
        """
        Tokenize a list of midi files and save the tokens to files
        Parameters
        ----------
        midi_files
        output_dir
        quantization
        fast_chord_inference
        chord_range

        Returns
        -------

        """
        for midi_file in tqdm(midi_files, 'Tokenizing files'):
            try:
                tokens = self.tokenize_midi_file(midi_file, quantization=quantization, fast_chord_inference=fast_chord_inference, chord_range=chord_range)
                filename = os.path.basename(midi_file).replace('.mid', '.txt')
                output_file = os.path.join(output_dir, filename)
                with open(output_file, 'w') as f:
                    f.write(sep.join(tokens))
            except Exception as e:
                if allow_error:
                    print(f"Error while tokenizing {midi_file}: {e}")
                else:
                    raise e


    def worker(self, args):
        gc.collect()  # Force garbage collection at the start of the worker
        midi_file, output_dir, quantization, fast_chord_inference, interceptors, chord_range, sep, allow_error = args
        try:
            tokens = self.tokenize_midi_file(midi_file, quantization=quantization,
                                             fast_chord_inference=fast_chord_inference,
                                             chord_range=chord_range,
                                             interceptors=interceptors
                                             )
            filename = os.path.basename(midi_file).replace('.mid', '.txt')
            output_file = os.path.join(output_dir, filename)
            with open(output_file, 'w') as f:
                f.write(sep.join(tokens))
        except Exception as e:
            if allow_error:
                print(f"Error while tokenizing {midi_file}: {e}")
            else:
                raise e
        finally:
            gc.collect()  # Force garbage collection at the end of the worker

    def tokenize_to_files(self, midi_files, output_dir, quantization=8,
                          fast_chord_inference=True, chord_range=None, sep="\n",
                          allow_error=True, interceptors=None, num_processes=4
                          ):
        args = [(midi_file, output_dir, quantization, fast_chord_inference, interceptors, chord_range, sep, allow_error)
                for midi_file in midi_files]

        with Pool(num_processes, maxtasksperchild=1) as pool:
            for _ in tqdm(pool.imap_unordered(self.worker, args), total=len(midi_files), desc='Tokenizing files'):
                pass


    def density_to_density_str(self, density):
        # densities = {'low': (0, 0.55), 'medium': (0.55, 1.5), 'high': (1, 2), 'very_high': (2, 10)}
        if density < 0.55:
            return 'low'
        elif density < 1.9:
            return 'medium'
        elif density < 3:
            return 'high'
        else:
            return 'very_high'

    def tokenize_to_bytes(self, score, as_one_str=True):
        tokens = self.tokenize(score)
        return self.tokens_to_bytes(tokens, as_one_str=as_one_str)

    def untokenize_from_bytes(self, bytes):
        tokens = self.bytes_to_tokens(bytes, to_str=False)
        return self.untokenize(tokens)


    def tokenize(self, score, only_chords=False, for_prompt=False):
        if self.dict['options'].get('random_instrument_permutation', False):
            import random
            instruments = score.instruments
            random.shuffle(instruments)
            sort_dict = {ins: idx for idx, ins in enumerate(instruments)}

        if isinstance(score, Chord):
            score = Score(chords=[score])
        tokens = []
        for idx, chord in enumerate(score):
            densities = chord.to_score().extract_densities()
            octave_means = chord.to_score().extract_mean_octaves()
            amplitude_means = chord.to_score().extract_mean_amplitudes()

            tokens_chord = self.tokenize_chord(chord, only_chords=only_chords)
            tokens += tokens_chord
            if self.dict['options'].get('next_chord_token', False) and not only_chords:
                # Last element
                if idx < len(score.chords) - 1:
                    next_chord = score.chords[idx + 1]
                    tokens_next_chord = self.tokenize_next_chord(next_chord)
                    tokens += tokens_next_chord
                else:
                    if self.dict['options'].get('will_end_token', True):
                        tokens += [self.WILL_END]

            if not only_chords:

                if self.dict['options'].get('random_instrument_permutation', False):
                    items = sorted(chord.score.items(), key=lambda x: sort_dict[x[0]])
                else:
                    items = chord.score.items()

                for ins, melody in items:
                    ins_name, ins_part = ins.split('__')
                    tokens_ins_name = self.INSTRUMENT_NAME + '__' + ins_name
                    tokens.append(tokens_ins_name)
                    if self.dict['options'].get('voice_token', True):
                        tokens_ins_part = self.INSTRUMENT_PART + '__' + ins_part
                        tokens.append(tokens_ins_part)
                    if self.dict['options'].get('density_token', False):
                        instrument_density = densities[ins]
                        density_str = self.density_to_density_str(instrument_density)
                        tokens_density = self.DENSITY + '__' + density_str
                        tokens += [tokens_density]
                    if self.dict['options'].get('average_octave_token', False):
                        tokens_average_octave = self.AVERAGE_OCTAVE + '__' + str(octave_means[ins])
                        tokens += [tokens_average_octave]
                    if self.dict['options'].get('amplitude_token', False):
                        tokens_amplitude = self.AMPLITUDE + '__' + str(amplitude_means[ins])
                        tokens += [tokens_amplitude]
                    for note in melody:
                        tokens_note = self.tokenize_note(note)
                        tokens += tokens_note
                    if self.dict['options'].get('melody_end_token', False):
                        tokens.append(self.MELODY_END)

        if self.dict['options'].get('end_token', True):
            tokens.append(self.END)

        return tokens

    def tokenize_chord_duration(self, chord_duration):
        token_chord_duration_num = self.CHORD_DURATION_NUM + '__' + str(chord_duration.numerator)
        token_chord_duration_den = self.CHORD_DURATION_DEN + '__' + str(chord_duration.denominator)
        return [token_chord_duration_num, token_chord_duration_den]


    def get_avalaible_denominators(self):
        if self.tokenizer is not None:
            words = self.tokenizer.vocab
            self.denoms = [int(w.split('__')[1]) for w in words if 'NOTE_DURATION_DEN' in w]
        return self.denoms

    def tokenize_note(self, note):
        note_type = self.NOTE_TYPE + '__' + note.type
        note_degree = self.NOTE_VAL + '__' + str(note.val)
        note_octave = self.NOTE_OCTAVE + '__' + str(note.octave)
        note_amp = self.NOTE_AMP + '__' + note.amp_figure

        # Limit denominator of duration to 4
        note_duration = frac(note.duration).limit_denominator(NOTE_DURATION_MAX_DENOMINATOR)
        available_denominators = self.denoms
        note_duration_den = min(available_denominators, key=lambda x: abs(note_duration.denominator - x))
        if note_duration.numerator == 0:
            return []

        # Find best approximation using denominator
        note_duration = frac(note_duration.numerator, note_duration_den)
        note_duration_num = self.NOTE_DURATION_NUM + '__' + str(note_duration.numerator)
        note_duration_den = self.NOTE_DURATION_DEN + '__' + str(note_duration.denominator)
        # Create the list
        tokens = [note_type]
        if (note.type not in ['r', 'l']) or not self.dict['options'].get('silence_continuation_eco', False):
            tokens += [note_degree, note_octave, note_amp]

        tokens += [note_duration_num, note_duration_den]

        return tokens

    def tokenize_chord(self, chord, only_chords=False):
        tokens = []
        if self.dict['options']['chord_change_token']:
            tokens.append(self.CHORD_CHANGE)

        chord_degree = self.CHORD_DEGREE + '__' + str(chord.element)
        chord_octave = self.CHORD_OCTAVE + '__' + str(chord.full_octave)
        chord_extension = self.CHORD_EXTENSION + '__' + str(chord.extension)
        tonality_degree = self.TONALITY_DEGREE + '__' + str(chord.tonality.degree)
        tonality_mode = self.TONALITY_MODE + '__' + chord.tonality.mode

        # Create the list
        tokens += [chord_degree, tonality_degree, tonality_mode, chord_octave]

        if self.dict['options'].get('chord_extension_token', False):
            tokens += [chord_extension]

        if self.dict['options'].get('chord_duration_token', False) and not only_chords:
            chord_duration = frac(chord.duration).limit_denominator(CHORD_DURATION_MAX_DENOMINATOR)
            chord_duration_num = self.CHORD_DURATION_NUM + '__' + str(chord_duration.numerator)
            chord_duration_den = self.CHORD_DURATION_DEN + '__' + str(chord_duration.denominator)
            tokens += [chord_duration_num, chord_duration_den]

        return tokens

    def tokenize_next_chord(self, chord):
        tokens = []

        chord_degree = self.NEXT_CHORD_DEGREE + '__' + str(chord.element)
        chord_octave = self.NEXT_CHORD_OCTAVE + '__' + str(chord.octave)
        tonality_degree = self.NEXT_TONALITY_DEGREE + '__' + str(chord.tonality.degree)
        tonality_mode = self.NEXT_TONALITY_MODE + '__' + chord.tonality.mode

        tokens += [chord_degree, tonality_degree, tonality_mode, chord_octave]

        if self.dict['options'].get('next_chord_duration_token', True):
            chord_duration = frac(chord.duration).limit_denominator(CHORD_DURATION_MAX_DENOMINATOR)
            chord_duration_num = self.NEXT_CHORD_DURATION_NUM + '__' + str(chord_duration.numerator)
            chord_duration_den = self.NEXT_CHORD_DURATION_DEN + '__' + str(chord_duration.denominator)
            tokens += [chord_duration_num, chord_duration_den]

        if self.dict['options']['chord_extension_token']:
            chord_extension = self.NEXT_CHORD_EXTENSION + '__' + str(chord.extension)
            tokens += [chord_extension]

        return tokens





    def untokenize(self, tokens):
        if isinstance(tokens, str):
            # Split by \n or whitespace
            tokens = tokens.split()

        score = Score()
        current_chord = None
        current_melody = None
        current_instrument_name = None
        current_instrument_part = None

        current_chord_duration_num = None
        current_chord_duration_den = None
        chord_duration = None
        current_melody_duration = 0
        note_duration_num = 0
        note_duration_den = 0
        note_val = 0
        note_type = 'r'
        note_octave = 0
        note_amp = 'mf'

        current_instrument_idx = {}

        for token in tokens:
            # Split token into key and value
            if token in [self.END, self.CHORD_CHANGE, self.WILL_END, self.SCORE_START]:
                continue

            if token == self.MELODY_END:
                # Check if melody duration is equal to chord duration, else add a rest
                if current_melody_duration < chord_duration:
                    delta = chord_duration - current_melody_duration
                    note_duration_num = int(delta.numerator)
                    note_duration_den = int(delta.denominator)
                    note_duration = frac(note_duration_num, note_duration_den)
                    if note_duration > 0:
                        note = Note(type='r', val=0, octave=0, duration=note_duration)
                        current_melody.notes.append(note)
                # Then continue
                current_chord.score[current_instrument_name + '__' + current_instrument_part] = current_melody
                continue

            try:
                key, value = token.split('__')
            except:
                raise ValueError(f"Invalid token {token}")

            if key == self.CHORD_DEGREE:
                if current_chord is not None:
                    score.chords.append(current_chord)
                current_chord = Chord(element=int(value), tonality=Tonality(0, 'M'))
                current_instrument_idx = {}

            elif key == self.CHORD_OCTAVE:
                current_chord.octave = int(value)

            elif key == self.CHORD_DURATION_NUM:
                current_chord_duration_num = int(value)

            elif key == self.CHORD_DURATION_DEN:
                current_chord_duration_den = int(value)
                chord_duration = frac(current_chord_duration_num, current_chord_duration_den)

            elif key == self.TONALITY_DEGREE:
                current_chord.tonality.degree = int(value)  # Assuming Tonality can be constructed from a string

            elif key == self.TONALITY_MODE:
                current_chord.tonality.mode = value

            elif key == self.CHORD_EXTENSION:
                current_chord.extension = value

            elif key == self.INSTRUMENT_NAME:
                current_instrument_name = value
                current_melody_duration = 0
                note_duration_num = 0
                note_duration_den = 0
                current_melody = Melody(notes=[])

                if current_instrument_name not in current_instrument_idx:
                    current_instrument_idx[current_instrument_name] = 0
                else:
                    current_instrument_idx[current_instrument_name] += 1

                if not self.dict['options'].get('voice_token', False):
                    current_instrument_part = str(current_instrument_idx[current_instrument_name])

            elif key == self.INSTRUMENT_PART:
                # Assuming that instrument part is not used directly in Melody
                current_instrument_part = value

            elif key == self.NOTE_TYPE:
                note_type = value

            elif key == self.NOTE_VAL:
                note_val = int(value)

            elif key == self.NOTE_OCTAVE:
                note_octave = int(value)

            elif key == self.NOTE_AMP:
                note_amp = value

            elif key == self.NOTE_DURATION_NUM:
                note_duration_num = int(value)

            elif key == self.NOTE_DURATION_DEN:
                note_duration_den = int(value)
                current_note_duration = frac(note_duration_num, note_duration_den)
                if current_melody_duration + current_note_duration > chord_duration:
                    delta = current_melody_duration + current_note_duration - chord_duration
                    note_duration_num = int(delta.numerator)
                    note_duration_den = int(delta.denominator)
                    note_duration = current_note_duration - frac(note_duration_num, note_duration_den)
                    if note_duration > 0:
                        note = Note(type=note_type, val=note_val, octave=note_octave, duration=note_duration)
                        note = note.set_amp(note_amp)
                        current_melody.notes.append(note)
                    current_melody_duration += note_duration
                    current_chord.score[current_instrument_name + '__' + current_instrument_part] = current_melody
                else:
                    note = Note(type=note_type, val=note_val, octave=note_octave, duration=current_note_duration)
                    note = note.set_amp(note_amp)
                    current_melody_duration += current_note_duration
                    current_melody.notes.append(note)

                current_instrument = current_instrument_name + '__' + current_instrument_part
                current_chord.score[current_instrument] = current_melody

        # Add the last chord to the score
        if current_chord is not None:
            score.chords.append(current_chord)

        return score


import os
import glob

current_file_path = os.path.abspath(__file__)
current_file_dir = os.path.dirname(current_file_path)
midi_files = glob.glob(os.path.join(current_file_dir, "corpus", "*.mid"))


def list_corpus():
    return [os.path.basename(m).split('.')[0] for m in midi_files]


def get_midi_path_from_corpus(name):
    return os.path.join(current_file_dir, "corpus", f"{name}.mid")




import re


class BPEIterator:
    r"""
    An iterable class to be used when training a musiclang tokenizer with BPE.

    It loads tokens text files to be used with the Hugging Face
    tokenizers library to build a vocabulary with BPE.

    It splits the tokens into different sequences using all the control tokens as separator
    Eg : CHORD_CHANGE INSTRUMENT_NAME__piano INSTRUMENT_PART__0 NOTE_TYPE_s NOTE_VAL__0 NOTE_OCTAVE__0 ... will be splitted as
    ['CHORD_CHANGE', 'INSTRUMENT_NAME__piano', 'INSTRUMENT_PART__0', 'NOTE_TYPE_s NOTE_VAL__0 NOTE_OCTAVE__0 ...']
    separating note tokens from control tokens.

    """

    def __init__(self, tokenizer, files_paths, control_tokens=[]) -> None:
        self.tokenizer = tokenizer
        self.files_paths = files_paths
        self.control_tokens = control_tokens
        self.__iter_count = 0

    def load_file(self, path):
        """
        Load a MIDI file and convert it to its byte representation.

        :param path: path to the file to load.
        :return: the byte representation of the file.
        """
        with open(path, 'r') as f:
            text = f.read()

        # list of str (bytes)
        bytes_ = self.tokenizer.tokens_to_bytes(text)
        bytes_ = bytes_[:8000]

        # Split
        split_pattern = '|'.join([re.escape(token) for token in self.control_tokens])
        bytes_ = re.split(f'({split_pattern})', bytes_)
        bytes_ = [b for b in bytes_ if len(b) > 0]
        return bytes_

    def __len__(self):
        """
        Return the number of files in the training corpus.

        :return: number of files in the training corpus.
        """
        return len(self.files_paths)

    def __getitem__(self, idx: int):
        """
        Convert the ``idx``th file to its byte representation.

        :param idx: idx of the file to convert.
        :return: byte representation of the file.
        """
        return self.load_file(self.files_paths[idx])

    def __iter__(self):  # noqa:D105
        return self

    def __next__(self) :  # noqa:D105
        if self.__iter_count >= len(self):
            self.__iter_count = 0
            raise StopIteration

        self.__iter_count += 1
        return self[self.__iter_count - 1]

    def __str__(self):
        """
        Return the ``str`` representation of the iterator.

        :return: string description.
        """
        return f"{self.tokenizer} - {len(self)} files"




from .chord_tokenizer import ChordTokenizer, ChordDetokenizer
from .tokenizer import MusicLangTokenizer
from .bpe_tokenizer import MusicLangBPETokenizer
from .template_extractor import midi_file_to_template, score_to_template



import setuptools
from setuptools import setup, Extension, find_packages
from setuptools.command.install import install
from setuptools.command.install_lib import install_lib
from setuptools.command.build_clib import build_clib
from setuptools.command.build_py import build_py
from setuptools.command.build_ext import build_ext
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info
import subprocess
import os

def custom_command():
    subprocess.check_call(['make', '-C', 'musiclang_predict/c/'])

# Define the C extension
class CustomInstall(install):
    """Custom build command that runs a Makefile."""

    def run(self):
        print('is called install')
        custom_command()
        # Call the superclass methods to handle Python extension building, if any
        install.run(self)

class CustomInstallClib(build_clib):
    """Custom build command that runs a Makefile."""

    def run(self):
        print('is called install')
        custom_command()
        # Call the superclass methods to handle Python extension building, if any
        build_clib.run(self)

class CustomInstallExt(install_lib):
    """Custom build command that runs a Makefile."""

    def run(self):
        print('is called install')
        custom_command()
        # Call the superclass methods to handle Python extension building, if any
        install_lib.run(self)

class CustomInstallBuildPy(build_py):
    """Custom build command that runs a Makefile."""

    def run(self):
        print('is called install')
        custom_command()
        # Call the superclass methods to handle Python extension building, if any
        build_py.run(self)


class CustomEggInfoCommand(egg_info):
    def run(self):
        print('is called egg')
        custom_command()
        egg_info.run(self)


module = Extension('musiclang_predict.c',
                   sources=['musiclang_predict/c/run.c'],
                   include_dirs=[],
                   extra_compile_args=['-Ofast', '-fPIC', '-shared'])

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding='utf-8')

setuptools.setup(
    name="musiclang-predict",
    version="1.1.8",
    author="Florian GARDIN",
    author_email="fgardin.pro@gmail.com",
    description=("Controllable symbolic music generation with generative AI"
                ),
    cmdclass={
                'build_py': CustomInstallBuildPy,
              },
    #ext_modules=[module],
    long_description=long_description,
    long_description_content_type="text/markdown",
    project_urls={
        'Documentation': 'https://github.com/MusicLang/musiclang_predict',
        'Source': 'https://github.com/MusicLang/musiclang_predict',
        'Tracker': 'https://github.com/MusicLang/musiclang_predict/issues',
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
            "musiclang>=0.25",
            "torch",
            "transformers",
            "tokenizers",
            "torchtoolkit",
            "accelerate"
                      ],
    packages=setuptools.find_packages(include='*'),
    package_data={'musiclang_predict': ['c/*.h', 'c/*.c', 'c/*.so', 'c/*.dll', 'c/Makefile', 'corpus/*.mid'],
                  },
    include_package_data=True,
    python_requires=">=3.6",
)